# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict
from typing import List

import click
from vdk.internal.control.utils import cli_utils
from vdk.internal.control.utils import output_printer
from vdk.plugin.data_sources.config import ConfigClassMetadata
from vdk.plugin.data_sources.factory import SingletonDataSourceFactory


def list_data_sources() -> List[Dict]:
    result = []
    for source in SingletonDataSourceFactory().list():
        item = {
            "name": source.name,
            "description": ConfigClassMetadata(source.config_class).get_description(),
        }

        result.append(item)
    return result


def list_config_options(data_source_name: str) -> List[Dict]:
    result = []
    sources = [
        s for s in SingletonDataSourceFactory().list() if s.name == data_source_name
    ]
    if sources:
        source = sources[0]
        for field in ConfigClassMetadata(source.config_class).get_config_fields():
            result.append(
                dict(
                    name=field.name(),
                    description=field.description()
                    + (
                        f" (default: {field.default()})"
                        if field.default() is not None
                        else ""
                    ),
                )
            )
    return result


# TODO using ConfigClassMetadata fields you can create prompt to automatically ask config options
# and generate data flow file out of that. `vdk sources --interactive`


@click.command(
    name="data-sources",
    help="Explore VDK data sources ."
    """This command all registered data sources

     Example:
     # To list all data sources, run:
     `vdk data-sources --list`

     # To show config options for give data source
     vdk data-sources --config csv
     """,
)
@click.option(
    "--list", is_flag=True, help="List registered data sources that can be used. "
)
@click.option(
    "--config",
    type=click.STRING,
    help="List configuration options for the data source ",
)
@click.pass_context
@cli_utils.output_option()
def data_sources(ctx: click.Context, list: bool, config: str, output: str):
    if list:
        listed_sources = list_data_sources()
        output_printer.create_printer(output).print_table(listed_sources)
    if config:
        output_printer.create_printer(output).print_table(list_config_options(config))



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from dataclasses import MISSING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar


def config_class(name: str, description: str, **kwargs):
    """
    A decorator to mark a class as a configuration class. It extends dataclass decorator with config related options.
    Each field in a decorator class must be defined with config_field

    Example::

        @config_class(name="group", description="Just example")
        ExampleConfig:
            option1: int = config_field(description="Option 1")
            option2: bool = config_filed(description="Option 2")

    :param name: The name of the configuration group representing by the class.
    :param description: The description of the configuration group.
    :param kwargs: Additional arguments passed same as arguments of :func:`dataclass`.
    """

    def decorator(cls):
        cls = dataclass(cls, **kwargs)

        setattr(cls, "config_group_name", name)
        setattr(cls, "config_group_description", description)

        for field_name, field_info in cls.__dataclass_fields__.items():
            if "config_description" not in field_info.metadata:
                raise TypeError(
                    f"The field '{field_name}' must be declared using config_field."
                )

        return cls

    return decorator


def config_field(
    *,
    description: str,
    is_sensitive: bool = False,
    default=MISSING,
    init=True,
    repr=True,
    hash=None,
    compare=True,
    metadata=None,
    **kwargs,
):
    """
    Define a field within a configuration class. This is extending dataclass.field with addition parameters

    :param description: Description of the configuration field.
    :param is_sensitive: Indicates if the configuration field contains sensitive information.
    The rest are the same as in dataclass.field()
    """

    if metadata is None:
        metadata = {}
    metadata.update(
        {"config_description": description, "config_is_sensitive": is_sensitive}
    )

    return field(
        **kwargs,
        default=default,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


class ConfigFieldMetadata:
    """
    Metadata class for configuration fields.

    Provides methods to access the metadata of a specific configuration field.
    """

    def __init__(self, field_class, field_name: str):
        self._field_name = field_name
        self._field_class = field_class

    def __get_field_metadata_value(self, metadata_key: str) -> Any:
        return self._field_class.__dataclass_fields__[self._field_name].metadata.get(
            metadata_key, None
        )

    def name(self) -> str:
        return self._field_name

    def description(self) -> str:
        return self.__get_field_metadata_value("config_description")

    def is_sensitive(self) -> bool:
        return self.__get_field_metadata_value("config_is_sensitive")

    def default(self) -> Optional[Any]:
        default_value = self._field_class.__dataclass_fields__[self._field_name].default
        return default_value if default_value != MISSING else None


class ConfigClassMetadata:
    """
    Metadata class for configuration classes.

    Provides methods to access the metadata of a specific configuration class.
    """

    def __init__(self, cls: Type):
        self._config_cls: Type[config_class] = cls
        self.__check_config_class()

    def __check_config_class(self):
        if not hasattr(self._config_cls, "config_group_description"):
            raise ValueError(
                f"cls {self._config_cls} is not decorated with @config_class."
            )

    def get_group_name(self) -> str:
        return getattr(self._config_cls, "config_group_name")

    def get_description(self) -> str:
        return getattr(self._config_cls, "config_group_description")

    def get_config_fields(self) -> List[ConfigFieldMetadata]:
        fields_meta = []
        for field_info in fields(self._config_cls):
            fields_meta.append(ConfigFieldMetadata(self._config_cls, field_info.name))
        return fields_meta


T = TypeVar("T")


def is_config_class(cls: Type):
    return ConfigClassMetadata(cls) is not None


def create_config_from_dict(config_class: Type[T], config_data: Dict[str, Any]) -> T:
    field_types = {f.name: f.type for f in fields(config_class)}
    processed_data = {}

    for key, value in config_data.items():
        if hasattr(field_types[key], "__dataclass_fields__"):
            # Nested dataclass
            processed_data[key] = create_config_from_dict(field_types[key], value)
        else:
            processed_data[key] = value

    return config_class(**processed_data)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict
from typing import TypeVar

import toml
from vdk.plugin.data_sources.mapping.definitions import DataFlowMappingDefinition
from vdk.plugin.data_sources.mapping.definitions import Definitions
from vdk.plugin.data_sources.mapping.definitions import DestinationDefinition
from vdk.plugin.data_sources.mapping.definitions import SourceDefinition

T = TypeVar("T")


def definitions_from_dict(data: Dict) -> Definitions:
    sources = {
        k: SourceDefinition(id=k, **v) for k, v in data.get("sources", {}).items()
    }
    destinations = {
        k: DestinationDefinition(id=k, **v)
        for k, v in data.get("destinations", {}).items()
    }
    flows_list = []
    for flow in data.get("flows", []):
        if flow["from"] not in sources:
            raise ValueError(
                f"Data source {flow['from']} is not defined and cannot be used in a flow.\n"
                f"Please define it first in the [sources] section of the configuration.\n"
                f"Currenty defined surces: {list(sources.keys())}"
            )
        if flow["to"] not in destinations:
            raise ValueError(
                f"Data destination {flow['to']} is not defined and cannot be used in a flow.\n"
                f"Please define it first in the [destinations] section of the configuration.\n"
                f"Currenty defined destinations: {list(destinations.keys())}"
            )

        flow_obj = DataFlowMappingDefinition(
            from_source=sources[flow["from"]], to_destination=destinations[flow["to"]]
        )
        flows_list.append(flow_obj)

    return Definitions(sources=sources, destinations=destinations, flows=flows_list)


def load_config(file_path: str) -> Definitions:
    parsed_toml = toml.load(file_path)
    return definitions_from_dict(parsed_toml)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import typing
from abc import abstractmethod
from typing import Any
from typing import Dict

from vdk.api.job_input import IProperties


class IDataSourceStateStorage:
    @abstractmethod
    def read(self, data_source: str) -> Dict[str, typing.Any]:
        """
        Read the state from underlying storage
        :param data_source: the data source name
        :return:
        """
        pass

    @abstractmethod
    def write(self, data_source: str, state: Dict[str, typing.Any]):
        """
        write (persist) the current state of the data source to underlying data storage.
        :param data_source:
        :param state:
        :return:
        """
        pass


class IDataSourceState:
    @abstractmethod
    def read_stream(self, stream_name: str) -> Dict[str, typing.Any]:
        """
        Read the state of a stream within a data source
        :param stream_name:
        :return:
        """

    @abstractmethod
    def update_stream(self, stream_name: str, state: Dict[str, typing.Any]):
        """
        Update the state of a stream within the datasource
        :param stream_name:
        :param state:
        :return:
        """
        pass

    @abstractmethod
    def read_others(self, key: str) -> Dict[str, typing.Any]:
        """
        Read state of a data source that is not related to streams
        :param key:
        :return:
        """
        pass

    @abstractmethod
    def update_others(self, key: str, state: Dict[str, typing.Any]):
        """
        Update state of a data source not related to streams
        :param key:
        :param state:
        :return:
        """
        pass


class DataSourceState(IDataSourceState):
    def __init__(self, state_storage: IDataSourceStateStorage, source: str):
        self._state_storage = state_storage
        self._source = source

    def __read_source(self) -> Dict[str, typing.Any]:
        source_state = self._state_storage.read(self._source)
        return source_state if source_state is not None else {}

    def read_stream(self, stream_name: str) -> Dict[str, typing.Any]:
        return self.__read_source().get("streams", {}).get(stream_name, {})

    def read_others(self, key: str) -> Dict[str, typing.Any]:
        return self.__read_source().get("others", {}).get(key, {})

    def update_stream(self, stream_name: str, state: Dict[str, typing.Any]):
        source_state = self.__read_source()
        source_state.setdefault("streams", {})[stream_name] = state
        self._state_storage.write(self._source, source_state)

    def update_others(self, key: str, state: Dict[str, typing.Any]):
        source_state = self.__read_source()
        source_state.setdefault("others", {})[key] = state
        self._state_storage.write(self._source, source_state)


class InMemoryDataSourceStateStorage(IDataSourceStateStorage):
    """
    In memory state useful for testing purposes
    """

    def __init__(self):
        self._state = {}

    def read(self, data_source: str) -> Dict[str, Any]:
        return self._state.get(data_source, {})

    def write(self, data_source: str, state: Dict[str, Any]):
        self._state[data_source] = state


class PropertiesBasedDataSourceStorage(IDataSourceStateStorage):
    KEY = ".vdk.data_sources.state"

    def __init__(self, properties: IProperties):
        self._properties = properties

    def read(self, data_source: str) -> Dict[str, Any]:
        return self._properties.get_property(
            PropertiesBasedDataSourceStorage.KEY, {}
        ).get(data_source, {})

    def write(self, data_source: str, state: Dict[str, Any]):
        all_properties = self._properties.get_all_properties()
        all_properties.setdefault(PropertiesBasedDataSourceStorage.KEY, {})[
            data_source
        ] = state
        self._properties.set_all_properties(all_properties)


class DataSourceStateFactory:
    def __init__(self, storage: IDataSourceStateStorage):
        self._storage = storage

    def get_data_source_state(self, source: str) -> IDataSourceState:
        return DataSourceState(self._storage, source)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

import pytest
from vdk.plugin.data_sources import config
from vdk.plugin.data_sources.config import config_class
from vdk.plugin.data_sources.config import config_field
from vdk.plugin.data_sources.config import ConfigClassMetadata


def test_config_class():
    @config_class("test_name", "test_description")
    class TestConfig:
        pass

    assert hasattr(TestConfig, "config_group_name")
    assert TestConfig.config_group_name == "test_name"
    assert hasattr(TestConfig, "config_group_description")
    assert TestConfig.config_group_description == "test_description"


def test_config_field_without_default():
    @config_class("test_name", "test_description")
    class TestConfig:
        attr: int = config_field(description="attr_description")

    fields = ConfigClassMetadata(TestConfig).get_config_fields()
    assert len(fields) == 1
    metadata = fields[0]
    assert metadata.name() == "attr"
    assert metadata.description() == "attr_description"
    assert metadata.is_sensitive() is False
    assert metadata.default() is None


def test_config_field_with_default():
    @config_class("test_name", "test_description")
    class TestConfig:
        attr: int = config_field(description="attr_description", default=42)

    instance = TestConfig()
    assert instance.attr == 42

    instance = TestConfig(1)
    assert instance.attr == 1

    metadata = ConfigClassMetadata(TestConfig).get_config_fields()[0]
    assert metadata.name() == "attr"
    assert metadata.default() == 42


def test_config_field_sensitive():
    @config_class("test_name", "test_description")
    class TestConfig:
        attr: bool = config_field(description="attr_description", is_sensitive=True)

    metadata = ConfigClassMetadata(TestConfig).get_config_fields()[0]
    assert metadata.name() == "attr"
    assert metadata.is_sensitive() is True


def test_invalid_field_missing_config_field():
    with pytest.raises(TypeError):

        @config_class("test_name", "test_description")
        class TestConfig:
            attr: bool


def test_deserialize():
    @dataclass
    class SubTest:
        id: str
        num: int

    @config_class("name", "description")
    class Test:
        name: str = config_field(description="foo")
        sub_test: SubTest = config_field(description="subtest")

    actual_instance: Test = config.create_config_from_dict(
        Test, dict(name="my-name", sub_test=dict(id="my-id", num=12))
    )
    assert actual_instance.name == "my-name"
    assert actual_instance.sub_test.id == "my-id"
    assert actual_instance.sub_test.num == 12



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from copy import deepcopy
from typing import Any
from typing import Optional

import pytest
from vdk.api.job_input import IIngester
from vdk.api.job_input import IProperties
from vdk.plugin.data_sources.auto_generated import (
    AutoGeneratedDataSource,
)
from vdk.plugin.data_sources.auto_generated import (
    AutoGeneratedDataSourceConfiguration,
)
from vdk.plugin.data_sources.ingester import DataSourceIngester


class MockJobInput(IIngester, IProperties):
    def __init__(self):
        self.ingested_data = []
        self.props = {}

    def send_object_for_ingestion(
        self,
        payload: dict,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str],
        collection_id: Optional[str] = None,
    ):
        self.ingested_data.append(payload)

    def send_tabular_data_for_ingestion(
        self,
        rows: iter,
        column_names: list,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str],
        collection_id: Optional[str] = None,
    ):
        raise NotImplemented("send_tabular_data_for_ingestion")

    def get_property(self, name: str, default_value: Any = None) -> str:
        return self.props.get(name, default_value)

    def get_all_properties(self) -> dict:
        return self.props

    def set_all_properties(self, properties: dict):
        self.props = deepcopy(properties)


@pytest.mark.parametrize("num_records, num_streams", [(5, 2), (10, 1)])
def test_data_source_ingester(num_records, num_streams):
    # Arrange
    data_source, data_source_ingester, mock_job_input = arrange(
        num_records, num_streams
    )

    # Act
    data_source_ingester.start_ingestion("id", data_source)
    data_source_ingester.terminate_and_wait_to_finish()

    # Assert
    assert num_records * num_streams == len(mock_job_input.ingested_data)
    assert min(mock_job_input.ingested_data, key=lambda x: x["id"])["id"] == 1
    assert max(mock_job_input.ingested_data, key=lambda x: x["id"])["id"] == num_records


@pytest.mark.parametrize("num_records, num_streams", [(2, 2)])
def test_data_source_ingester_with_duplicate_source_raises_error(
    num_records, num_streams
):
    # Arrange
    data_source, data_source_ingester, mock_job_input = arrange(
        num_records, num_streams
    )

    # Act
    data_source_ingester.ingest_data_source("auto", data_source)
    with pytest.raises(ValueError):
        data_source_ingester.ingest_data_source(
            "auto", data_source, method="second_method"
        )


@pytest.mark.parametrize("num_records, num_streams", [(5, 2), (10, 1)])
def test_data_source_ingester_with_two_sources(num_records, num_streams):
    # Arrange
    data_source, data_source_ingester, mock_job_input = arrange(
        num_records, num_streams
    )

    # Act
    data_source_ingester.ingest_data_source("auto", data_source)
    data_source_ingester.ingest_data_source(
        "auto2", data_source, method="second_method"
    )
    data_source_ingester.terminate_and_wait_to_finish()

    # Assert
    assert num_records * num_streams * 2 == len(mock_job_input.ingested_data)


@pytest.mark.parametrize("num_records, num_streams", [(5, 2), (10, 1)])
def test_data_source_ingester_with_state(num_records, num_streams):
    # First time
    mock_job_input = MockJobInput()

    data_source_ingester = DataSourceIngester(mock_job_input)
    data_source = get_data_source(num_records, num_streams)

    data_source_ingester.ingest_data_source("auto", data_source)
    data_source_ingester.terminate_and_wait_to_finish()

    # Second time:

    data_source_ingester = DataSourceIngester(mock_job_input)
    data_source = get_data_source(num_records, num_streams)

    data_source_ingester.ingest_data_source("auto", data_source)
    data_source_ingester.terminate_and_wait_to_finish()

    # Assert
    assert num_records * num_streams * 2 == len(mock_job_input.ingested_data)
    assert min(mock_job_input.ingested_data, key=lambda x: x["id"])["id"] == 1
    assert (
        max(mock_job_input.ingested_data, key=lambda x: x["id"])["id"]
        == num_records * 2
    )


def arrange(num_records, num_streams):
    mock_job_input = MockJobInput()
    data_source_ingester = DataSourceIngester(mock_job_input)
    data_source = get_data_source(num_records, num_streams)
    return data_source, data_source_ingester, mock_job_input


def get_data_source(num_records, num_streams):
    config = AutoGeneratedDataSourceConfiguration(
        num_records=num_records, include_metadata=False, num_streams=num_streams
    )
    data_source = AutoGeneratedDataSource()
    data_source.configure(config)
    return data_source



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import queue
import traceback
from dataclasses import dataclass
from queue import Queue
from threading import Thread
from typing import Callable
from typing import cast
from typing import List
from typing import Optional

from vdk.api.job_input import IIngester
from vdk.api.job_input import IJobInput
from vdk.plugin.data_sources.data_source import DataSourceError
from vdk.plugin.data_sources.data_source import DataSourcePayload
from vdk.plugin.data_sources.data_source import (
    DataSourcesAggregatedException,
)
from vdk.plugin.data_sources.data_source import IDataSource
from vdk.plugin.data_sources.data_source import (
    IDataSourceErrorCallback,
)
from vdk.plugin.data_sources.data_source import IDataSourceStream
from vdk.plugin.data_sources.data_source import (
    RetryDataSourceStream,
)
from vdk.plugin.data_sources.data_source import (
    StopDataSourceStream,
)
from vdk.plugin.data_sources.state import DataSourceStateFactory
from vdk.plugin.data_sources.state import PropertiesBasedDataSourceStorage

log = logging.getLogger(__name__)


@dataclass
class IngestDestination:
    destination_table: Optional[str] = None
    method: Optional[str] = None
    target: Optional[str] = None
    collection_id: Optional[str] = None
    map_function: Optional[
        Callable[[DataSourcePayload], Optional[DataSourcePayload]]
    ] = None


@dataclass
class IngestQueueEntry:
    data_source_id: str
    stream: IDataSourceStream
    destinations: List[IngestDestination]
    error_callback: Optional[IDataSourceErrorCallback] = None


class DataSourceIngester:
    def __init__(self, job_input: IJobInput):
        self.__ingestion_queue = Queue()
        self.__actual_ingester = cast(IIngester, job_input)
        self.__worker_threads = self._start_workers(8)
        self.__being_ingested_streams = set()
        self.__stored_exceptions = queue.SimpleQueue()
        self.__state_factory = DataSourceStateFactory(
            PropertiesBasedDataSourceStorage(job_input)
        )

    def _start_workers(self, number_of_worker_threads: int) -> List[Thread]:
        threads = []
        for i in range(number_of_worker_threads):
            thread_name = f"data-source-ingest-{i}"
            t = Thread(target=self._process_ingestion_queue, name=thread_name)
            t.daemon = True
            t.start()
            threads.append(t)
        return threads

    def _process_ingestion_queue(self):
        while True:
            ingest_entry: IngestQueueEntry = self.__ingestion_queue.get()
            if self._is_termination_signal(ingest_entry):
                self.__ingestion_queue.task_done()
                break
            try:
                self._ingest_stream(ingest_entry)
            except StopDataSourceStream:
                log.debug(f"Stopping data source stream {ingest_entry.stream.name()}")
            except RetryDataSourceStream:
                log.debug(
                    f"Retrying to ingest data source stream {ingest_entry.stream.name()}"
                )
                self.__ingestion_queue.put(ingest_entry)
            except BaseException as e:
                log.exception("Ingestion failed")
                if ingest_entry.error_callback:
                    self._handle_exception(e, ingest_entry)
                else:
                    tb = traceback.extract_tb(e.__traceback__)
                    # Print only the last part of the traceback
                    last_trace = tb[-1]
                    filename = os.path.basename(last_trace.filename)
                    last_trace_message = (
                        f"{filename}:{last_trace.lineno} {last_trace.name}"
                    )

                    error_message = f"Failed to ingest stream {ingest_entry.stream.name()} with error {e} in {last_trace_message}"
                    log.warning(error_message)
                    self.__stored_exceptions.put((ingest_entry.stream.name(), e))
            finally:
                self.__ingestion_queue.task_done()

    def _handle_exception(self, e, ingest_entry):
        try:
            ingest_entry.error_callback(
                DataSourceError(data_stream=ingest_entry.stream, exception=e)
            )
        except StopDataSourceStream:
            log.debug(f"Stopping data source stream {ingest_entry.stream.name()}")
        except RetryDataSourceStream:
            log.debug(
                f"Retrying to ingest data source stream {ingest_entry.stream.name()}"
            )
            self.__ingestion_queue.put(ingest_entry)
        except BaseException as callback_exception:
            error_message = (
                f"Failed to ingest stream {ingest_entry.stream.name()} with error {e}."
                f" And error callback failed as well with error {callback_exception} "
            )
            log.warning(error_message)
            self.__stored_exceptions.put((ingest_entry.stream.name(), e))

    @staticmethod
    def _is_termination_signal(queue_item):
        return queue_item is None

    def _ingest_stream(self, ingest_entry: IngestQueueEntry):
        for payload in ingest_entry.stream.read():
            log.debug(f"Ingest payload {payload}")
            if not payload.data and not payload.state:
                log.warning(
                    f"{ingest_entry.stream.name()} returned payload without any data or state. "
                    f"This is pretty meaningless so we will skip it. But it is suspicious."
                )
                continue

            if payload.data:
                for destination in ingest_entry.destinations:
                    destination_table = self._infer_destination_table(
                        ingest_entry, destination, payload
                    )

                    payload_to_sent = DataSourcePayload(
                        payload.data, payload.metadata, payload.state, destination_table
                    )
                    if destination.map_function:
                        payload_to_sent = destination.map_function(payload_to_sent)
                        if not payload_to_sent:
                            continue

                    self.__actual_ingester.send_object_for_ingestion(
                        payload=payload_to_sent.data,
                        destination_table=payload_to_sent.destination_table,
                        method=destination.method,
                        target=destination.target,
                        collection_id=destination.collection_id,
                    )

            if payload.state:
                # TODO: it would be better to update the state after the actual ingester
                # is finished. So it might be good idea to add a callback so that when send object for ingestion
                # finishes successfully then state is updated.
                data_source_state = self.__state_factory.get_data_source_state(
                    ingest_entry.data_source_id
                )
                data_source_state.update_stream(
                    ingest_entry.stream.name(), payload.state
                )

    @staticmethod
    def _infer_destination_table(
        ingest_entry: IngestQueueEntry, destination: IngestDestination, payload
    ):
        if destination.destination_table:
            destination_table = destination.destination_table
        elif payload.destination_table:
            destination_table = payload.destination_table
        else:
            destination_table = ingest_entry.stream.name()
        return destination_table

    def start_ingestion(
        self,
        data_source_id: str,
        data_source: IDataSource,
        destinations: List[IngestDestination] = None,
        error_callback: Optional[IDataSourceErrorCallback] = None,
    ):
        if data_source_id not in self.__being_ingested_streams:
            self.__being_ingested_streams.add(data_source_id)
        else:
            raise ValueError(
                f"Data source is already in ingestion queue. "
                f"Ingestion same data source more than once is currently not supported."
            )
        if not destinations:
            destinations = [IngestDestination()]

        data_source.connect(self.__state_factory.get_data_source_state(data_source_id))

        for stream in data_source.streams():
            entry = IngestQueueEntry(
                data_source_id, stream, destinations, error_callback
            )
            self.__ingestion_queue.put(entry)

    def ingest_data_source(
        self,
        data_source_id: str,
        data_source: IDataSource,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
        error_callback: Optional[IDataSourceErrorCallback] = None,
    ):
        self.start_ingestion(
            data_source_id,
            data_source,
            [IngestDestination(destination_table, method, target, collection_id)],
            error_callback,
        )

    def terminate_and_wait_to_finish(self):
        self.__ingestion_queue.join()
        self.__ingestion_queue.put(None)

    def raise_on_error(self):
        if not self.__stored_exceptions.empty():
            collected_exceptions = dict()
            while not self.__stored_exceptions.empty():
                stream_name, exception = self.__stored_exceptions.get()
                collected_exceptions[stream_name] = exception

            # Raise a custom exception that encapsulates all the stored exceptions
            raise DataSourcesAggregatedException(collected_exceptions)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.plugin.data_sources.state import DataSourceState
from vdk.plugin.data_sources.state import IDataSourceState
from vdk.plugin.data_sources.state import InMemoryDataSourceStateStorage


def __data_source_state(source) -> IDataSourceState:
    storage = InMemoryDataSourceStateStorage()
    return DataSourceState(state_storage=storage, source=source)


def test_get_empty_data_source():
    assert __data_source_state("non_existent_source").read_stream("foo") == {}
    assert __data_source_state("non_existent_source").read_others("foo") == {}


def test_update_and_get_data_stream_state():
    data_source_name = "source1"
    data_stream_name = "stream1"
    new_state = {"key": "value"}

    data_source_state = __data_source_state(data_source_name)
    data_source_state.update_stream("stream", new_state)
    data_source_state.update_others("key", new_state)

    assert data_source_state.read_stream("stream") == new_state
    assert data_source_state.read_others("key") == new_state

    data_source_state.update_stream("stream", {"key": "value2"})
    data_source_state.update_others("key", {"key": "value2"})

    assert data_source_state.read_stream("stream") == {"key": "value2"}
    assert data_source_state.read_others("key") == {"key": "value2"}



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from vdk.plugin.data_sources.config import is_config_class
from vdk.plugin.data_sources.state import IDataSourceState


class IDataSourceConfiguration:
    """
    Interface representing the configuration for a data source.

    It's part of a definition of a data source - what configuraiton it requires.
    You need to implement a class and decoreated with @config_class decorator like this:

    Example::

    # Configuration for StackOverflow data source to interact with its API and extract data.
    @config_class(name="stackoverflow-api", description="Extract data from StackOverflow API")
    class StackOverflowDataSourceConfiguration(IDataSourceConfiguration):
        api_url: str = config_field(description="StackOverflow API URL")
        api_key: Optional[str] = config_field(description="API key", default="")
        api_endpoints: List[str] = config_field(
            description="List of API endpoints", default=['answers', 'questions', 'posts']
        )
        ...

    """

    def __new__(cls, *args, **kwargs):
        new_class = super().__new__(cls)
        if not is_config_class(new_class):
            raise TypeError(f"Class {cls} must be decorated with @config_class")
        return new_class


@dataclass(frozen=True)
class DataSourcePayload:
    """
    Encapsulates a single payload to be ingested coming from a data source.

    :param data: The data to be ingested.
    :param metadata: Additional metadata about the data.
    :param state: Optional state related to the data (for the timestamp or id of the payload) which will be peristed once the payload is sucessfully ingested
    """

    data: Optional[Dict[str, Any]]
    metadata: Optional[Dict[str, Union[int, str, bool, float, datetime]]]
    state: Optional[Dict[str, Any]] = field(default_factory=dict)
    destination_table: Optional[str] = None


class StopDataSourceStream(Exception):
    """Signal the end of a stream and there's no more data"""


class RetryDataSourceStream(Exception):
    """Signal the stream ingestion should be retried"""


class IDataSourceStream:
    """
    Abstract class for a Data Source Stream, representing a channel or resource to read data.

    For example each table in a database could be a stream.
    Or each separate API endpoint for an API could be a stream and so on.

    :Example::

        # Implement the data extraction from StackOverflow API
        class StackOverflowDataSourceStream(IDataSourceStream):

            def __init__(self, endpoint_url: str):
                self._endpoint_url = endpoint_url

            def read(self) -> Iterator[DataSourcePayload]:
                for item in stackoverflow_client.get(self._endpoint_url):
                    yield DataSourcePayload(item)

    """

    @abstractmethod
    def name(self) -> str:
        """
        :return: unique (within the data souce) name of the data source stream which can be used as a key and find by it.
        """
        pass

    @abstractmethod
    def read(self) -> Iterable[DataSourcePayload]:
        """
        Generator method or Iterator for reading data from the stream.

        :return: An iterable of DataSourcePayload objects.

        """
        pass


class IDataSource:
    """
    Abstract class for a Data Source, responsible for managing the connection and providing data streams.

    :Example::

        # Let's build data source for Stack Overflow API. We will split each API endpoint data into separate stream

        @data_source(name="stackoverflow-api",
                     config_class=StackOverflowDataSourceConfiguration)
        class StackOverflowDataSource(IDataSource):

            def configure(self, config: IDataSourceConfiguration):
                self._config = config

            def connect(self, state: IDataSourceState):
                if not self._streams:
                    self._streams = [StackOverflowDataSourceStream(endpoint) for endpoint in self._config.api_endpoints]

            def streams(self) -> List[IDataSourceStream]:
                return self._streams

            def disconnect(self):
                close_all_streams(self._streams)
                self._streams = []

    """

    @abstractmethod
    def configure(self, config: IDataSourceConfiguration):
        """
        Configure the data source.

        :param config: Data source configuration object.
            In datasource subclass you are encourages to use the specific subclass config as type hint.
        """

    @abstractmethod
    def connect(self, state: Optional[IDataSourceState]):
        """
        Establish a connection using provided configuration and last saved state.

        :param state: Data source state object.
        """

    @abstractmethod
    def disconnect(self):
        """
        Disconnect and clean up resources if needed.
        """

    @abstractmethod
    def streams(self) -> List[IDataSourceStream]:
        """
        Get the available streams for this data source.

        :return: List of IDataSourceStream objects.
        """


class DataSourcesAggregatedException(Exception):
    """
    Exception to aggregate multiple Data Sources failures.

    :Example::

        DataSourcesAggregatedException({"Stream1": Exception1, "Stream2": Exception2})
    """

    def __init__(self, data_streams_exceptions: Dict[str, Exception]):
        super().__init__(
            f"Data Sources failed to ingest data: {data_streams_exceptions}"
        )
        self.data_streams_exceptions = data_streams_exceptions


@dataclass
class DataSourceError:
    """
    Data class to encapsulate information about a Data Source ingestion error.

    :data_stream: The data stream where the error occurred.
    :exception: The exception that was raised.

    :Example::

        DataSourceError(data_stream=MyStream(), exception=SomeException())

    """

    data_stream: IDataSourceStream
    exception: BaseException


class IDataSourceErrorCallback:
    """
    Callback interface to be implemented for handling data source ingestion errors.

    Example::

        # can be a function
        def my_error_callback(error: DataSourceError):
            print(f"Stream: {error.data_stream.name()}, Exception: {error.exception}")

        # or class
        class MyErrorCallback(IDataSourceErrorCallback)

            def __call__(self, error: DataSourceError):
                print(f"Stream: {error.data_stream.name()}, Exception: {error.exception}")

    """

    @abstractmethod
    def __call__(self, error: DataSourceError):
        """
        Invoked when an error occurs during data ingestion.

        :param error:DataSourceError: Object containing details of the occurred error.

        :raises:
            StopDataSourceStream: Stops the current data stream without any errors
            RetryDataSourceStream: Retries ingesting the current data stream later
        """

    pass


class IDataMetricsBatchAnalyzer:
    """
    Implement metrics analyzer for given data source.
    This will be called for each payload.
    The implementation must provide constructor without arguments
    """

    def analyze_batch(self, payload: DataSourcePayload):
        """
        Analyze the payload as it's being collected.

        :param DataSourcePayload:
        :return:
        """


class IDataMetricsFullAnalyzer:
    """
    Implement metrics analyzer for given data source.
    This will be called after the data source is exhausted
    The implementation must provide constructor without arguments
    """

    def get_data_store_target(self) -> str:
        """
        This is the ingestion target where the data would be stored for further analysis
        """
        pass

    def get_data_store_method(self) -> str:
        """
        This is the ingestion method which will be used to store the data  for further analysis
        :return:
        """
        pass

    def analyze_at_the_end(self):
        """
        Does the actual analysis on the stored data.
        To use this you must also specify data store method and data store target.
        :return:
        """



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json

from click.testing import Result
from vdk.plugin.data_sources import plugin_entry
from vdk.plugin.test_utils.util_funcs import cli_assert_equal
from vdk.plugin.test_utils.util_funcs import CliEntryBasedTestRunner


def test_sources_command_list():
    runner = CliEntryBasedTestRunner(plugin_entry)

    result: Result = runner.invoke(["data-sources", "--list", "-o", "json"])

    cli_assert_equal(0, result)
    assert json.loads(result.stdout) == [
        {
            "description": "Configuration for Auto generated data source",
            "name": "auto-generated-data",
        }
    ]


def test_sources_command_config():
    runner = CliEntryBasedTestRunner(plugin_entry)

    result: Result = runner.invoke(
        ["data-sources", "--config", "auto-generated-data", "-o", "json"]
    )

    cli_assert_equal(0, result)
    assert len(json.loads(result.stdout)) == 3



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Collection
from typing import Dict
from typing import List
from typing import Type

from vdk.plugin.data_sources import config
from vdk.plugin.data_sources.data_source import IDataSource
from vdk.plugin.data_sources.data_source import (
    IDataSourceConfiguration,
)


def data_source(name: str, config_class: Type[IDataSourceConfiguration]):
    def inner_decorator(data_source_class: Type[IDataSource]):
        data_source_class.__data_source_name__ = name
        data_source_class.__data_source__config__class__ = config_class
        return data_source_class

    return inner_decorator


@dataclass
class DataSourceRegistryItem:
    """
    A registry item for Data Sources.

    :param name: The name identifier for the data source
    :param instance_class: The data source class that implements `IDataSource`
    :param config_class: The configuration class that implements `IDataSourceConfiguration`
    """

    name: str
    instance_class: Type[IDataSource]
    config_class: Type[IDataSourceConfiguration]


class DataSourceNotFoundException(Exception):
    def __int__(self, data_source_name: str, existing_data_source_names: List[str]):
        super().__init__(
            f"Data source with name {data_source_name} does not exists. "
            f"Existing data sources are {existing_data_source_names}"
        )

    pass


class IDataSourceFactory:
    @abstractmethod
    def register_data_source_class(
        self,
        data_source_class: Type[IDataSource],
    ):
        """
        Register a data source and its associated configuration class.

        :param data_source_class: The data source class that implements `IDataSource` and must be decoreated with @data_source

        Example::

        @data_source(name="csv", config_class=CsvDataSourceConfiguration)
        class CsvDataSource(IDataSource):
            ...
        """

    @abstractmethod
    def list(self) -> Collection[DataSourceRegistryItem]:
        """
        List all registered data sources.

        :return: A collection of registered data sources
        """

    @abstractmethod
    def create_data_source(self, name: str) -> IDataSource:
        """
        Create an instance of a registered data source.

        :param name: The name identifier for the data source
        :return: An instance of the data source
        """


class SingletonDataSourceFactory(IDataSourceFactory):
    """
    Singleton factory class to create and manage Data Source instances.
    The factory holds a registry of Data Source classes and their associated configuration classes.
    """

    _instance = None

    def __new__(cls):
        """
        Implement the Singleton pattern by creating a single instance or returning the existing one.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.__data_source_registry = {}
        return cls._instance

    def register_data_source_class(self, data_source_class: Type[IDataSource]):
        if not hasattr(data_source_class, "__data_source_name__") or not hasattr(
            data_source_class, "__data_source__config__class__"
        ):
            raise ValueError(
                "Invalid data_source_class."
                "data_source_class must extend IDataSource, and be decorated with @data_source."
            )

        name = getattr(data_source_class, "__data_source_name__")
        config_class = getattr(data_source_class, "__data_source__config__class__")
        self.__register_data_source(name, data_source_class, config_class)

    def __register_data_source(
        self,
        name: str,
        data_source_class: Type[IDataSource],
        data_source_config_class: Type[IDataSourceConfiguration],
    ):
        if not isinstance(data_source_class, type):
            raise ValueError(
                "data_source_class must a class definition. Not a class instance for example."
            )
        if not issubclass(data_source_class, IDataSource):
            raise ValueError("data_source_class must be a class inheriting IDataSource")
        if not isinstance(data_source_config_class, type):
            raise ValueError(
                "data_source_config_class must a class definition. Not a class instance for example."
            )
        if not issubclass(data_source_config_class, IDataSourceConfiguration):
            raise ValueError(
                "data_source_config_class must be a class inheriting IDataSourceConfiguration"
            )

        self.__data_source_registry[name] = DataSourceRegistryItem(
            name, data_source_class, data_source_config_class
        )

    def list(self) -> Collection[DataSourceRegistryItem]:
        return self.__data_source_registry.values()

    def create_data_source(self, name: str) -> IDataSource:
        registry_item = self.__get_source_item(name)

        instance = registry_item.instance_class()
        return instance

    def create_configuration(self, name: str, config_data: Dict[str, Any]):
        registry_item = self.__get_source_item(name)
        instance = config.create_config_from_dict(
            registry_item.config_class, config_data
        )
        return instance

    def __get_source_item(self, name) -> DataSourceRegistryItem:
        registry_item = self.__data_source_registry.get(name)
        if not registry_item:
            raise DataSourceNotFoundException(
                name, list(self.__data_source_registry.keys())
            )
        return registry_item



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from vdk.plugin.data_sources.data_source import IDataSource
from vdk.plugin.data_sources.data_source import IDataSourceConfiguration
from vdk.plugin.data_sources.factory import data_source
from vdk.plugin.data_sources.factory import DataSourceNotFoundException
from vdk.plugin.data_sources.factory import (
    SingletonDataSourceFactory,
)


class MockDataSourceConfig(Mock, IDataSourceConfiguration):
    pass


@data_source("mock", MockDataSourceConfig)
class MockDataSource(Mock, IDataSource):
    pass


def test_register_and_list_data_source():
    with patch(
        SingletonDataSourceFactory.__module__
        + "."
        + SingletonDataSourceFactory.__name__
        + "._instance",
        None,
    ):
        factory = SingletonDataSourceFactory()

        factory.register_data_source_class(MockDataSource)

        data_sources = factory.list()
        assert len(data_sources) == 1

        factory2 = SingletonDataSourceFactory()
        assert len(factory2.list()) == 1


def test_create_data_source():
    with patch(
        SingletonDataSourceFactory.__module__
        + "."
        + SingletonDataSourceFactory.__name__
        + "._instance",
        None,
    ):
        factory = SingletonDataSourceFactory()

        factory.register_data_source_class(MockDataSource)

        data_source_instance = factory.create_data_source("mock")
        assert data_source_instance is not None


def test_create_data_source_not_foound():
    with patch(
        SingletonDataSourceFactory.__module__
        + "."
        + SingletonDataSourceFactory.__name__
        + "._instance",
        None,
    ):
        factory = SingletonDataSourceFactory()

        with pytest.raises(DataSourceNotFoundException):
            factory.create_data_source("not-found")



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import List

from vdk.plugin.data_sources.auto_generated import (
    AutoGeneratedDataSource,
)
from vdk.plugin.data_sources.auto_generated import (
    AutoGeneratedDataSourceConfiguration,
)
from vdk.plugin.data_sources.auto_generated import (
    AutoGeneratedDataSourceStream,
)
from vdk.plugin.data_sources.data_source import DataSourcePayload
from vdk.plugin.data_sources.state import DataSourceStateFactory
from vdk.plugin.data_sources.state import InMemoryDataSourceStateStorage


def test_in_memory_data_source_stream():
    config = AutoGeneratedDataSourceConfiguration(
        num_records=3, include_metadata=True, num_streams=1
    )
    stream = AutoGeneratedDataSourceStream(config, 1)
    records: List[DataSourcePayload] = list(stream.read())
    assert len(records) == 3
    for record in records:
        assert "id" in record.data
        assert "name" in record.data
        assert "stream" in record.data
        assert "timestamp" in record.metadata
        assert isinstance(record.metadata["timestamp"], datetime)


def test_data_source_plugin():
    config = AutoGeneratedDataSourceConfiguration(
        num_records=3, include_metadata=True, num_streams=2
    )
    data_source = AutoGeneratedDataSource()

    data_source.configure(config)
    data_source.connect(
        DataSourceStateFactory(InMemoryDataSourceStateStorage()).get_data_source_state(
            "foo"
        )
    )
    streams = data_source.streams()
    assert len(streams) == 2

    for i, stream in enumerate(streams):
        records = list(stream.read())
        assert len(records) == 3
        for record in records:
            assert "id" in record.data
            assert "name" in record.data
            assert record.data["stream"] == i
            assert "timestamp" in record.metadata

    data_source.disconnect()
    assert len(data_source.streams()) == 0



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.job_input import IJobInput
from vdk.plugin.data_sources.factory import SingletonDataSourceFactory
from vdk.plugin.data_sources.ingester import DataSourceIngester
from vdk.plugin.data_sources.ingester import IngestDestination
from vdk.plugin.data_sources.mapping.definitions import DataFlowMappingDefinition
from vdk.plugin.data_sources.mapping.definitions import Definitions


class DataFlowInput:
    """
    Manages the data flow from source to destination using the provided job input.

    :param job_input: Instance of IJobInput, providing context for the current job.
    """

    def __init__(self, job_input: IJobInput):
        self._job_input = job_input
        self._data_source_ingester = DataSourceIngester(job_input)
        self._data_source_factory = SingletonDataSourceFactory()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Terminates and waits for data source ingestion to finish.
        """
        self._data_source_ingester.terminate_and_wait_to_finish()
        self._data_source_ingester.raise_on_error()

    def start(self, flow_definition: DataFlowMappingDefinition):
        """
        Start data flow from a specific source to a specific destination.

        :param flow_definition: The definition of the source and destination flow.
        """
        source = self._data_source_factory.create_data_source(
            flow_definition.from_source.name
        )
        source_config = self._data_source_factory.create_configuration(
            flow_definition.from_source.name, flow_definition.from_source.config
        )
        source.configure(source_config)

        destination = IngestDestination(
            method=flow_definition.to_destination.method,
            target=flow_definition.to_destination.target,
            map_function=flow_definition.map_function,
        )
        self._data_source_ingester.start_ingestion(
            flow_definition.from_source.id, source, [destination]
        )

    def start_flows(self, definitions: Definitions):
        """
        Start data flows based on a list of defined mappings.

        :param definitions: Definitions containing the data flow mappings.
        """
        for flowDefinition in definitions.flows:
            self.start(flowDefinition)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Iterator
from typing import List

from vdk.plugin.data_sources.config import config_class
from vdk.plugin.data_sources.config import config_field
from vdk.plugin.data_sources.data_source import DataSourcePayload
from vdk.plugin.data_sources.data_source import IDataSource
from vdk.plugin.data_sources.data_source import (
    IDataSourceConfiguration,
)
from vdk.plugin.data_sources.data_source import IDataSourceStream
from vdk.plugin.data_sources.factory import data_source
from vdk.plugin.data_sources.state import IDataSourceState


@config_class(
    name="auto_generated", description="Configuration for Auto generated data source"
)
class AutoGeneratedDataSourceConfiguration(IDataSourceConfiguration):
    num_records: int = config_field(
        description="Number of records to return", default=2
    )
    include_metadata: bool = config_field(
        description="If true autogenerated metadata is included in the response",
        default=False,
    )
    num_streams: int = config_field(
        description="The number of streams the data source would have", default=1
    )


class AutoGeneratedDataSourceStream(IDataSourceStream):
    """
    A single data stream in AutoGeneratedDataSource
    """

    def name(self) -> str:
        return f"stream_{self._stream_number}"

    def __init__(
        self,
        config: AutoGeneratedDataSourceConfiguration,
        stream_number: int,
        start_id: int = 0,
    ):
        self._config = config
        self._stream_number = stream_number
        self._data = self._generate_data(start_id)

    def _generate_data(self, start_id: int) -> List[DataSourcePayload]:
        generated_data = []
        for i in range(self._config.num_records):
            data = {
                "id": start_id + 1 + i,
                "name": f"Stream_{self._stream_number}_Name_{i}",
                "stream": self._stream_number,
            }
            metadata = (
                {"timestamp": datetime.now()} if self._config.include_metadata else {}
            )
            generated_data.append(
                DataSourcePayload(
                    data=data, metadata=metadata, state={"last_id": data["id"]}
                )
            )
        return generated_data

    def read(self) -> Iterator[DataSourcePayload]:
        for i in range(0, len(self._data), 1):
            yield self._data[i]


@data_source(
    name="auto-generated-data", config_class=AutoGeneratedDataSourceConfiguration
)
class AutoGeneratedDataSource(IDataSource):
    """
    Data source who is only generating some dummy data for testing purposes.
    """

    def __init__(self):
        self._config = None
        self._streams = []

    def configure(self, config: AutoGeneratedDataSourceConfiguration):
        self._config = config

    def connect(self, state: IDataSourceState):
        if not isinstance(self._config, AutoGeneratedDataSourceConfiguration):
            raise RuntimeError(
                f"config type must be {AutoGeneratedDataSourceConfiguration}"
            )
        self._streams = [
            AutoGeneratedDataSourceStream(
                self._config, i, state.read_stream(f"stream_{i}").get("last_id", 0)
            )
            for i in range(self._config.num_streams)
        ]

    def disconnect(self):
        self._streams = []

    def streams(self) -> List[IDataSourceStream]:
        return self._streams



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools

"""
Builds a package with the help of setuptools in order for this package to be imported in other projects
"""

__version__ = "0.1.1069988349"

setuptools.setup(
    name="vdk-data-sources",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="Enables Versatile Data Kit (VDK) to integrate with various data sources by providing a unified interface for data ingestion and management.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=[
        "vdk-core",
        "toml",
        "vdk-control-cli",
    ],  # vdk-control-cli is becasue of output printer
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    # This is the only vdk plugin specifc part
    # Define entry point called "vdk.plugin.run" with name of plugin and module to act as entry point.
    entry_points={
        "vdk.plugin.run": ["vdk-data-sources = vdk.plugin.data_sources.plugin_entry"]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    project_urls={
        "Documentation": "https://github.com/vmware/versatile-data-kit/tree/main/projects/vdk-plugins/vdk-data-sources",
        "Source Code": "https://github.com/vmware/versatile-data-kit/tree/main/projects/vdk-plugins/vdk-data-sources",
        "Bug Tracker": "https://github.com/vmware/versatile-data-kit/issues/new/choose",
    },
)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.core.config import ConfigurationBuilder
from vdk.plugin.data_sources.auto_generated import AutoGeneratedDataSource
from vdk.plugin.data_sources.config import ConfigClassMetadata
from vdk.plugin.data_sources.factory import IDataSourceFactory
from vdk.plugin.data_sources.factory import SingletonDataSourceFactory
from vdk.plugin.data_sources.hook_spec import DataSourcesHookSpec

"""
Include the plugins implementation. For example:
"""


class DataSourcesPlugin:
    @hookimpl
    def vdk_data_sources_register(self, data_source_factory: IDataSourceFactory):
        data_source_factory.register_data_source_class(AutoGeneratedDataSource)

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        # define the config options so they show up in the help.
        for ds in SingletonDataSourceFactory().list():
            config_class_meta = ConfigClassMetadata(ds.config_class)
            for field in config_class_meta.get_config_fields():
                group_name = config_class_meta.get_group_name()
                config_builder.add(
                    key=f"a data source {group_name} config's option {field.name()}",
                    default_value=field.default(),
                    description=field.description(),
                    is_sensitive=field.is_sensitive(),
                )


# TODO: add ingest.toml type of job step which would automatically execute ingestion flows declared in TOML wihtout any python code


@hookimpl
def vdk_start(plugin_registry: IPluginRegistry, command_line_args: List):
    plugin_registry.add_hook_specs(DataSourcesHookSpec)
    plugin_registry.load_plugin_with_hooks_impl(
        DataSourcesPlugin(), "DataSourcesPlugin"
    )

    plugin_registry.hook().vdk_data_sources_register.call_historic(
        kwargs=dict(data_source_factory=SingletonDataSourceFactory())
    )


@hookimpl
def vdk_command_line(root_command: click.Group):
    from vdk.plugin.data_sources.sources_command import data_sources

    root_command.add_command(data_sources)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from vdk.plugin.data_sources.data_source import DataSourcePayload


@dataclass
class SourceDefinition:
    """
    SourceDefinition class that defines the metadata and configuration for a data source.

    :param id: User provided identifier for the source "instance" which can be used to refer to that source later
    :param name: The name of the source. Must be the same name as the DataSource has been registered with (see vdk data-sources --list)
    :param config: The configuration for the source. Refer to documented data source config options (see vdk data-sources --config).
    """

    id: str
    name: str
    config: Dict[str, Any] = field(default_factory=dict)
    # TODO might be good idea to include initial state for first run
    # initial_state: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DestinationDefinition:
    """
    Describes the attributes and methods for a data destination.

    :param id: User provided identifier for the destination "instance" which can be used to refer to that source later.
    :param method: The ingestion method. Refer to IIngester for more information .
    :param target: The ingestion target. Refer to IIngester for more information .
    """

    id: str
    method: str
    target: Optional[str] = None
    # TODO: config when destination targets support being passed configs


@dataclass
class DataFlowMappingDefinition:
    """
    Defines how data flows from a source to a destination and the optional transformation in between.

    :param from_source: The source definition from where data originates.
    :param to_destination: The destination definition where data will be sent.
    :param map_function: An optional callable to transform data as it moves from source to destination.
                It accepts a payload and returns a payload or None. Returning None means skippping the payload ingestion!
    """

    from_source: SourceDefinition
    to_destination: DestinationDefinition
    map_function: Optional[
        Callable[[DataSourcePayload], Optional[DataSourcePayload]]
    ] = None
    # TODO add configuration for mappings tables and columns or function


@dataclass
class Definitions:
    """
    Aggregates all the source, destination, and flow definitions for an ingestion pipeline.

    :param sources: Dictionary of source definitions.
    :param destinations: Dictionary of destination definitions.
    :param flows: List of data flow mapping definitions.
    """

    sources: Dict[str, SourceDefinition]
    destinations: Dict[str, DestinationDefinition]
    flows: List[DataFlowMappingDefinition]



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookspec
from vdk.plugin.data_sources.factory import IDataSourceFactory


class DataSourcesHookSpec:
    @hookspec(historic=True)
    def vdk_data_sources_register(self, data_source_factory: IDataSourceFactory):
        """
        Register a data source and its associated configuration class.

        :param data_source_factory: the factory where the data sources should be registered. Provide those param:
            :name: The name identifier for the data source
            :data_source_class: The data source class that implements `IDataSource`. Only the type. Not an instance of it!

        Example::

            data_source_factory.register_data_source_class(data_source_class=CsvDataSource)

            #where CsvDataSource is like
            @data_source(name="csv", config_class=CsvDataSourceConfig)
            CsvDataSource(IDataSource):
               ...
        """
